program test_bicgstab
!-------------------------------------------------------------------------------------
! Test driver for a serial bicgstab solver in bicgstab.f90. This driver also  writes
! out a file results containing the history of the relative residual vector, so that
! it can plotted in Matlab with a short script. editing. For that reason, most of the
! performance data is printed to the file preceeded by the Matlab comment indicator
! "%". The parameter maxits is the maximal allowed iterations when calling bicgstab,
! and on return it has the actual number of iterations taken.
!
! The input variable printfreq specifies how many of the relative residual norms to
! print out for plotting. The bicgstab() subroutine computes all of them; printfreq
! only controls the output on return from it. If it is set to be <= 0, no residual
! norms are printed out at all.
!
! Integers are mostly declared as "int4", which are defined in the kindlytypes.f90
! module. "int8" integers are selected_int_kind(18), which means 64-bit integers. It's
! unlikely to need more than the 2 billion upper limit on the integers, but maybe in
! 2024 that will change.
!
! If only 32-bits are needed, change kindlytypes.f90, or change "int8" in all the
! files where it appears.
!
! benoisy.f90 is an include file with just one variable, that sets a level of
! debugging messages. benoisy < 1 gives no output to the screen, while benoisy = 3
! prints damned near everything. Way too many lines of this code are just to deal
! with the output specified by the variable benoisy.
!
! Interfaces are declared for all of the subprograms that have allocated arrays as
! dummy variables. Be sure to change those if the corresponding calling signatures
! are changed.
!
! The array timings is declared to have length 6, but only four timings need
! to be returned:
!       timings(1) = time spent in initialization
!       timings(2) = time spent in iteration
!       timings(3) = time spent in matrix-vector multiply bmux()
!       timings(4) = time spent in discretization
! The extra entries are for any unforeseen timings of interest. For all of the 
! timings, no repetition loop is needed. The problem sizes are large enough
! to allow even low resolution clocks to perform adequately. For bmux(), the
! times can be accumulated during the iteration, viz., initialize
!       timings[2] = 0.0;
! and then in the iteration phase, 
!       tbmux = *elapsedtime();
!       bmux(vals, rows, cols, nrows, ncols, nnz, x, y);
!       tbmux = *elapsedtime() - tbmux;
!       timings[2] += tbmux;
! Notice that bmux is called three places in the algorithm. You can include or
! exclude the time for the initialization invocation, since the number of
! iterations will be enough to make it account for less than 0.1% of the overall
! time.
!
! The basic logic/flow is:
!
!   open output file results
!   call getparameters to get matrix sizes and iteration control parameters
!   allocate matrix/vectors needed
!   call discretizer
!   call bicgstab
!   clean up the results file by appending Matlab instructions
!   ... ?
!   profit!
!   then go home 
!
!-----------------
! Randall Bramley
! Department of Computer Science
! Indiana University, Bloomington
!--------------------------------------------------------------------------------

    use kindlytypes, only: int4 => integer4, int8 => integer8, real8
    use iso_fortran_env, only: stdout => output_unit, &
                               stderr => error_unit

    use noisy
    implicit none

    ! Allocatables
    real(kind=real8), allocatable, dimension(:)   :: vals, x, b, relres
    integer(kind=int4), allocatable, dimension(:):: rows, cols

    ! Other junk and stuff
    real(kind=real8), external    :: elapsedtime
    real(kind=real8)    :: tstart, time_taken
    real(kind=real8)    :: timings(6)   ! Mostly the timings bicgstab returns
    real(kind=real8)    :: tolerance, flops
    integer(kind=int4)  :: nrows, ncols, nnz, job, k
    integer(kind=int4)  :: meshnodes, p, printfreq
    integer(kind=int4)  :: maxits, iflag
    logical:: printx  ! Whether or not print the solution vector at the end

    integer(kind=int4) :: resultsfile, iostatus, allocation_status
    character(len=15) :: date, time, zone ! Time stamp variables  

!==========
interface
!==========
    subroutine discretizer(meshpoints, nrows, ncols, nnz, rows, cols, vals, rhs, job)
        use kindlytypes, only: int4 => integer4, int8 => integer8, real8
        use noisy
        implicit none    
        integer(kind=int4):: meshpoints , nrows, ncols, nnz, row, col, job
        integer(kind=int4), allocatable:: rows(:), cols(:)
        real(kind=real8), allocatable :: vals(:), rhs(:)
    end subroutine discretizer

    subroutine bicgstab(vals,    & ! The double precision values in the matrix
                        rows,    & ! Usual row "pointers", length nrow+1
                        cols,    & ! The column indices, length nnz
                        nrow,    & ! Number of rows in the matrix
                        ncol,    & ! Number of columns in the matrix
                        nnz,     & ! Number of nonzeros in A
                        x,       & ! Solution vector
                        b,       & ! Right hand side vector
                        tol,     & ! Upper bound on relative residual
                        maxits,  & ! Maximum allowed nuber of iterations
                        relres,  & ! Relative residual for each iteration
                        timings, & ! Timings returned by BiCGstab
                        iflag)    ! Error flag
        use kindlytypes, only: int4 => integer4, int8 => integer8, real8
        implicit none
        integer(kind=int4) :: nrow, ncol, nnz
        integer(kind=int4) :: rows(nnz), cols(nnz)
        integer(kind=int4) :: iflag, maxits
        real(kind=real8) :: x(nrow), b(nrow), vals(nnz), tol, relres(maxits)
        real(kind=real8), dimension(:) :: timings
    end subroutine bicgstab

    subroutine getparameters(meshnodes, maxits, tolerance, printfreq, &
                             nrow, ncol, nnz, printx)
        use kindlytypes, only: int4 => integer4, int8 => integer8, real8
        use noisy
        implicit none
        logical :: printx
        real(kind=real8), intent(out) :: tolerance
        integer(kind=int4), intent(out) :: maxits, printfreq
        integer(kind=int4) :: meshnodes, nrow, ncol, nnz
    end subroutine getparameters

    subroutine timestamp(outunit, date, time, zone)
        use kindlytypes, only: int4 => integer4
        implicit none
        character(len=15), optional, intent(inout) :: date, time, zone
        integer(kind=int4), intent(in) :: outunit
    end subroutine timestamp

!==============
end interface
!==============

    ! OK, ready to actually start *doing* something.  
    if (benoisy > 0) write(stdout,200) 'opening results from driver'
    open(newunit=resultsfile, file='results', action='write', form="formatted", &
             iostat=iostatus )
    if (benoisy > 0) then
        write(stdout,500) 'results opened with LUN = ', resultsfile
        write(stdout,500) '      and with iostatus = ', iostatus
    end if

    ! Get the parameters 
    if (benoisy > 0) write(stdout,200) 'Calling getparameters.'
    call getparameters(meshnodes, maxits, tolerance, printfreq, &
                      nrows, ncols, nnz, printx)
    if (benoisy > 0) write(stdout,200) 'Back from getparameters'
    if (benoisy > 1) then
        write(stdout,700) 'After getparameters,    nrows  = ', nrows
        write(stdout,700) 'After getparameters,    ncols  = ', ncols
        write(stdout,700) 'After getparameters,     nnz   = ', nnz
        write(stdout,700) 'After getparameters,    maxits = ', maxits
        write(stdout,700) 'After getparameters, printfreq = ', printfreq
        write(stdout,600) 'After getparameters, tolerance = ', tolerance
    end if

    ! Get the sizes of arrays that need to be allocated
    if (benoisy > 0) write(stdout,200) 'Calling discretizer for sizes'
    job = 0
    timings(4) = elapsedtime()
    call discretizer(meshnodes, nrows, ncols, nnz, rows, cols, vals, b, job)
    timings(4) = elapsedtime() - timings(4)
    if (benoisy > 0) write(stdout,200) 'Back from discretizer'

    !============================================================================
    ! Allocate most of the memory required (bicgstab allocates automatic arrays)
    !============================================================================
    allocate(b(nrows),vals(nnz),rows(nnz),cols(nnz),x(nrows),stat=allocation_status)          
    if (allocation_status /= 0) then
        iflag = 1
        write(stderr,200) 'Out of memory when allocating A and x in main program'
        write(stderr,700) '    nrows = ', nrows
        write(stderr,700) '    ncols = ', ncols
        write(stderr,700) '    nnz = ', nnz
        stop
    end if
    allocate(relres(maxits+1), stat=allocation_status)          
    if (allocation_status /= 0) then
        iflag = 1
        write(stderr,200) 'Could not allocate relres vector'
        stop
    end if

    !========================================================
    ! Discretize the PDE and capture the time that it takes
    !========================================================
    if (benoisy > 0) write(stdout,200) 'Calling discretizer'
    job = 1
    timings(4) = elapsedtime()
    call discretizer(meshnodes, nrows, ncols, nnz, rows, cols, vals, b, job)
    timings(4) = elapsedtime() - timings(4)
    if (benoisy > 0) write(stdout,200) 'Back from discretizer'
    if (benoisy > 2) then
        write(stdout,5000) 'rows = ', rows
        write(stdout,5000) 'cols = ', cols
        write(stdout,6000) 'vals = ', vals
        write(stdout,6000) 'b = ', b
    end if

    !===================================================
    ! Call bicgstab, timing the overall time required 
    !===================================================
    if (benoisy > 0) write(stdout,200) 'Calling bicgstab'
    tstart = elapsedtime()
    x(1:nrows) = 0.0
    call bicgstab(vals,         & ! The double precision values in the matrix
                  rows,         & ! Usual row "pointers", length nrows+1
                  cols,         & ! The column indices, length nnz
                  nrows,        & ! Number of rows in the matrix
                  ncols,        & ! Number of columns in the matrix
                  nnz,          & ! Number of nonzeros in A
                  x,            & ! Solution vector
                  b,            & ! Right hand side vector
                  tolerance,    & ! Upper bound on relative residual
                  maxits,       & ! Maximum allowed nuber of iterations
                  relres,       & ! Convergence history of relative residuals
                  timings,      & ! Timings returned by BiCGstab
                  iflag)          ! Error flag
    time_taken = elapsedtime() - tstart
    if (benoisy > 0) write(stdout,200) 'Returned from bicgstab'
    if (benoisy > 0) write(stdout,700) 'iflag from bicgstab = ', iflag
    if (benoisy > 2) write(stdout,fmt=6000) 'x = ', x

    !=====================================
    ! Write results to the file results
    !=====================================
    flops = dble(2*nnz) + dble(3*nrows) + dble(maxits)*dble((2*nnz + 10*nrows))
    write(resultsfile,fmt=200) "% ---------------------------------------------------"
    write(resultsfile,fmt=200) "% All timings are given in seconds" 
    write(resultsfile,fmt=200) "% ---------------------------------"
    write(resultsfile,fmt=500) '% n = ', nrows, ' % matrix order'
    write(resultsfile,fmt=500) '% nnz = ', nnz, ' % number of nonzeros in the matrix'
    write(resultsfile,fmt=400) '% actual_iterations = ', maxits
    write(resultsfile,fmt=300) '% tolerance = ', tolerance
    write(resultsfile,fmt=400) '% iflag = ', iflag
    write(resultsfile,fmt=300) "% total_time          = ", time_taken
    write(resultsfile,fmt=300) "% discretization_time = ", timings(4)
    write(resultsfile,fmt=300) '% initialization_time = ', timings(1)
    write(resultsfile,fmt=300) '% iteration_time      = ', timings(2)
    write(resultsfile,fmt=300) '% matvec_time         = ', timings(3)
    write(resultsfile,fmt=300) "% gflop_rate =  ", 1.0e-9*flops/time_taken
    write(resultsfile,fmt=200) "% -------------- Residual History -------------------"
    if (printfreq > 0) then
        do k = 1, maxits
            if (mod(k, printfreq) == 0) then
                write(resultsfile, 100) k, relres(k)                                          
            end if
        end do
    end if
    call timestamp(resultsfile, date, time, zone)
    close(resultsfile)

    if (printx) call printplate(meshnodes, x)

    100 format(i8, es27.17)                                                  
    200 format(a)
    300 format(a, 3x, es14.6, ';')
    400 format(a, i11, ';')
    500 format(a, i11, a)
    600 format(a, 3x, es14.6)
    700 format(a, i0)
    5000 format(a, *(/,3x, i0))  ! infinite repeat descriptor
    6000 format(a, *(/,3x, es27.17))


    deallocate(vals, rows, cols, x, b, relres) 

    stop

end program test_bicgstab
