subroutine printplate(meshpoints, x) 
!----------------------------------------------------------------------------------
! Print out the solution to a file "plate". This should only be done for smallish
! problem sizes. You can view the thing using the Matlab script showplate.m . The
! vector x is assumed to have the node values in natural order: from left to right,
! going from bottom to top.
!
! This is more complicated than expected, because the matlab script showplot.m does
! not set up for mesh or surf plots correctly unless the values are listed in a
! specific order. 
!----------------
! Randall Bramley
! Department of Computer Science
! Indiana University, Bloomington
!----------------------------------------------------------------------------------

    use iso_fortran_env, only: stderr => error_unit
    use kindlytypes, only: real8
    use coeffs
    implicit none

    integer :: meshpoints 
    real(kind=real8) :: x(1:meshpoints*meshpoints), h, val
    real(kind=real8), parameter :: zero = 0.0_real8, one = 1.0_real8
    integer :: ix, iy, indx, plate, errflag

    open(newunit=plate, file="plate", action="write", iostat=errflag)
    if (errflag /= 0) then
        write(stderr,777) 'printplate: unable to open file plate; flag = ', errflag
        stop
    end if

    h = 1.0_real8/(1.0_real8 + meshpoints)

    ! 1. print out bottom boundary condition values ( y = 0 )
    do ix = 0, meshpoints+1
        val = ix*h
        write(plate,888) ix, 0, BC(val, zero)
    end do

    ! 2. Now handle the rows between top and bottom boundary conditions. To make 
    ! everything in order, this is done in three phases: first the left most
    ! point (which is a BC), then the middle points on the row, and finally the
    ! right most point on the row (which is another BC).
    indx = 0
    do iy = 1, meshpoints
        val = iy*h
        ! left hand boundary for the row
        write(plate,888) 0, iy, BC(zero, val)
        ! interior nodes for the row
        do ix = 1, meshpoints
            indx = indx+1
            write(plate,888) ix, iy, x(indx)
            ! flush plate
        end do
        ! right hand boundary for the row
        write(plate,888) meshpoints+1, iy, BC(one, val)
    end do

    ! 3. print out top boundary condition values ( y = 1 )
    do ix = 0, meshpoints+1
        val = ix*h
        write(plate,888) ix, meshpoints+1, BC(val, one)
    end do

    close(plate)

    777 format(a, i0)
    888 format(i0, 3x, i0, 5x, es25.17)

end subroutine printplate
