%--------------------------------------------------------------------------------
% This Matlab script will slurp in the values for the temperatures on a plate
% discretized onto a (N+2) x (N+2) mesh. The outermost edges are all assumed
% to be boundary conditions, and this script can either plot just the  boundary
% conditions, or the full plate. If the full plate is shown, the boundary
% conditions are also displayed but using different colors.
%
% If you want to just see the boundary conditions, set
%       showBConly = true;
% and the other Booleans to false.
%
% The input file "plate" should have two integers and a double on each row. The
% integers give the (i,j) coordinates of the corresponding value, where indexing
% goes from 0 to N+1. For the mesh plot, you must have the entries "in order",
% including the boundary values. The subroutine printplate.f90 does that.
%
% The boundary conditions are plotted using a different color and marker for the
% four sides, to help them stand out from the interior nodes when N is large. That
% is not done for the mesh and surfl plots because I don't know how.
%
% Don't use this for large meshes; this is primarily provided for debugging and
% visualization.
%--------------------------------------------------------------------------------

% Control parameters for what to show.
showBConly = false;
showall    = true; 
showmesh   = true;
showsurf   = true;

T = load('plate');
i    = T(:,1); 
j    = T(:,2);
vals = T(:,3);
N = max(i)-1;     % Mesh node numbers s go from 0 to N+1

% Need to put the boundary conditions in order for the mesh plot. First, sort
% based on the first (x) coordinate:
[i, IX] = sort(i);
j       = j(IX);
vals    = vals(IX);

clear T           % Don't need it around anymore
pos = [0.1  0.1    0.7   0.7];

% Plot the whole damned thing :
if showall 
    fighandle = figure('Units', 'normalized', 'Position', pos);
    plot3(i, j, vals, 'b+')
    xlabel('x')
    ylabel('y')
    zlabel('Temperature')
    title('Plate Temperature Values')
end

if (showBConly || showall)
    % First check the boundary conditions alone
    If = find(i == 0);
    Il = find(i == max(i));

    Jf = find(j == 0);
    Jl = find(j == max(j));

    if showBConly
        fighandle = figure('Units', 'normalized', 'Position', pos);
        xlabel('Bottom')
        ylabel('Left-hand size')
        zlabel('Temperature')
        title(['Boundary nodes of plate: N = ' num2str(N)])
    else
        hold on
    end
    plot3(i(If), j(If), vals(If), 'c+', ...     % cyan plus mark
          i(Il), j(Il), vals(Il), 'r+', ...     % red plus mark
          i(Jf), j(Jf), vals(Jf), 'g*', ...     % green asterisk
          i(Jl), j(Jl), vals(Jl), 'mo')         % magenta o mark
    hold off
end

if (showmesh  || showsurf)
    % Need to turn the 1D array of vals into a 2D array G for the mesh plot. Plus 
    % need to keep in mind that Matlab requires array indices to start from 1.
    [X, Y] = meshgrid(0:1:N+1, 0:1:N+1);
    G = zeros(N+2, N+2);
    for iy = 0:N+1
        for ix = 0:N+1
            G(ix+1, iy+1) = vals(ix+1 + iy*(N+2));
        end
    end

    if showmesh
        fighandle = figure('Units', 'normalized', 'Position', pos);
        mesh(X, Y, G)
        % For looking at the log of the values do 
        % H = G;
        % H(H<=0) = NaN;
        % mesh(X, Y, log(H))
        title('Computed Solution u(x,y)');
        zlabel('Temperature')
        xlabel('x')
        ylabel('y')
    end

    if showsurf
        fighandle = figure('Units', 'normalized', 'Position', pos);
        surfl(X,Y,G)
        shading interp
        title('Computed Solution u(x,y)');
        zlabel('Temperature')
        xlabel('x')
        ylabel('y')
    end

end
